"""Pro Custodibus Agent."""

from importlib_metadata import version

__version__ = version(__package__)

DISPLAY_NAME = "Pro Custodibus Agent"
DESCRIPTION = "Synchronizes your WireGuard settings with Pro Custodibus."
SERVICE_NAME = "ProCustodibusAgent"

DEFAULT_API_URL = "https://api.custodib.us"
DEFAULT_APP_URL = "https://pro.custodib.us"
DOCS_URL = "https://docs.procustodibus.com"



"""WireGuard utilities."""

import re
from base64 import b64decode, b64encode
from hashlib import sha256
from pathlib import Path
from platform import system
from subprocess import PIPE, run

from nacl.encoding import Base64Encoder
from nacl.public import PrivateKey


def get_wg_version(cnf):
    """Returns WireGuard client version.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Version string (eg '1.2.3').
    """
    if system() == "Windows":
        return _get_windows_wg_version(cnf)
    return _get_standard_wg_version(cnf)


def _get_standard_wg_version(cnf):
    """Returns standard WireGuard client version.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Version string (eg '1.2.3').
    """
    result = run([cnf.wg, "--version"], stdout=PIPE, check=False)  # noqa: S603
    output = result.stdout.decode("utf-8")
    return re.search(r"\sv(\S+)", output)[1]


def _get_windows_wg_version(cnf):
    """Returns Windows WireGuard client version.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Version string (eg '1.2.3').
    """
    exe = _get_windows_wireguard_exe(cnf)
    escaped = re.sub(r"[/\\]", r"\\\\", str(exe))
    try:
        args = [f"(Get-Item -path '{escaped}').VersionInfo.ProductVersion"]
        result = run(["powershell", *args], stdout=PIPE, check=False)  # noqa: S603 S607
        return result.stdout.decode("utf-8").strip()
    except OSError:
        args = ["datafile", "where", f"name='{escaped}'", "get", "version", "/value"]
        result = run(["wmic", *args], stdout=PIPE, check=False)  # noqa: S603 S607
        output = result.stdout.decode("utf-8").strip()
        return re.sub(r"Version=", "", output)


def _get_windows_wireguard_exe(cnf):
    """Returns path to Windows WireGuard client.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Path: Path to WireGuard executable.
    """
    wg = Path(cnf.wg)
    if cnf.wiresock:
        return wg
    directory = wg.parent
    if not directory.name:
        directory = Path("C:/Program Files/WireGuard/")
    return Path(directory, "wireguard.exe")


def run_wg_set(cnf, args):
    """Runs `wg set` command.

    Arguments:
        cnf (Config): Config object.
        args: Command arguments.
    """
    if not cnf.wiresock:
        run([cnf.wg, "set", *args], check=False)  # noqa: S603


def run_wg_show(cnf):
    """Runs `wg show` command.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Output of `wg show` command.
    """
    if cnf.wiresock:
        return ""
    result = run([cnf.wg, "show", "all", "dump"], stdout=PIPE, check=False)  # noqa: S603
    return result.stdout.decode("utf-8")


def parse_wg_show(s):
    """Parses `wg show` command output into a dict.

    Raises:
        ValueError: Output can't be parsed.

    Arguments:
        s (str): Output of `wg show` command.

    Returns:
        dict: Output of `wg show` command as a dict.
    """
    result = {}

    for line in s.splitlines():
        fields = line.split()
        len_fields = len(fields)

        if len_fields == 0:
            pass
        elif len_fields == 5:
            _add_interface(result, fields)
        elif len_fields == 9:
            _add_peer(result, fields)
        else:
            raise ValueError("can't parse line from wg show", line)

    return result


def annotate_wg_show(dst, src):
    """Copies properties from source interfaces to destination interfaces.

    Arguments:
        dst (str): Output of `wg show` command as dict.
        src (str): Output of `wg show` command as dict.

    Returns:
        dict: Dst dict updated with interface properties.
    """
    _default_all_endpoints_unavailable(dst)
    for name, src_iface in src.items():
        dst_iface = dst.get(name)
        if dst_iface:
            _annotate_wg_show_interface(dst_iface, src_iface)
    return dst


def _default_all_endpoints_unavailable(interfaces):
    """Defaults interfaces to down and peers unavailable.

    Arguments:
        interfaces (str): Output of `wg show` command as dict.

    Returns:
        dict: Dict updated with interface properties.
    """
    for iface in interfaces.values():
        iface["up"] = False
        peers = iface.get("peers")
        if peers:
            for peer in peers.values():
                peer["available"] = False


def _annotate_wg_show_interface(dst, src):
    """Copies properties from source interface to destination interface.

    Arguments:
        dst (str): One interface from output of `wg show` command as dict.
        src (str): One interface from output of `wg show` command as dict.

    Returns:
        dict: Dst dict updated with interface properties.
    """
    dst_peers = dst.get("peers")
    for name, src_value in src.items():
        if name == "peers" and dst_peers:
            _annotate_wg_show_peers(dst_peers, src_value)
        else:
            dst[name] = src_value
    return dst


def _annotate_wg_show_peers(dst, src):
    """Copies properties from source peer to destination peer.

    Arguments:
        dst (str): One peer from output of `wg show` command as dict.
        src (str): One peer from output of `wg show` command as dict.

    Returns:
        dict: Dst dict updated with peer properties.
    """
    for pubkey, src_peer in src.items():
        dst_peer = dst.get(pubkey)
        if dst_peer:
            dst_peer.update(src_peer)
        else:
            dst[pubkey] = src_peer


def update_socket_mark(data, cnf):
    """Updates the active socket mark with a new mark from the parsed interfaces.

    Arguments:
        data (dict): Dict parsed from `wg show` command.
        cnf (Config): Config object.

    Returns:
        dict: Same dict.
    """
    # don't override explicitly configured mark
    if cnf.fw_mark:
        return data
    fwmark = _get_fwmark_for_default_peer(data)
    # reset mark consumers if changed
    if cnf.socket_mark != fwmark:
        cnf.socket_mark = fwmark
        cnf.resolver = None
        cnf.transport = None
    return data


def filter_wg_show(data, cnf):
    """Removes configured elements from parsed output of `wg show`.

    Arguments:
        data (dict): Dict parsed from `wg show` command.
        cnf (Config): Config object.

    Returns:
        dict: Same dict with some items removed.
    """
    for interface in cnf.unmanaged_interfaces:
        data.pop(interface, None)
    if cnf.redact_secrets:
        data = remove_secrets_from_wg_show(data)
    elif cnf.redact_psk:
        data = remove_secrets_from_wg_show(data, psk_only=True)
    return data


def remove_secrets_from_wg_show(data, *, psk_only=False):
    """Removes secrets from parsed output of `wg show`.

    Arguments:
        data (dict): Dict parsed from `wg show` command.
        psk_only (bool): True to remove only preshared keys from output (default false).

    Returns:
        dict: Same dict with secret items removed.
    """
    for interface in data.values():
        if not psk_only:
            interface.pop("private_key", None)
        for peer in interface["peers"].values():
            peer.pop("preshared_key", None)
    return data


def _add_interface(result, fields):
    """Adds the the specified interface fields to the result.

    Arguments:
        result (dict): Result dict.
        fields (list): List of field values.
    """
    interface, private, public, port, fwmark = fields
    result[interface] = {
        "up": True,
        "private_key": private if private != "(none)" else "",
        "public_key": public if public != "(none)" else "",
        "listen_port": int(port),
        "fwmark": int(fwmark, 0) if fwmark != "off" else 0,
        "peers": {},
    }


def _add_peer(result, fields):
    """Adds the the specified peer fields to the result.

    Arguments:
        result (dict): Result dict.
        fields (list): List of field values.
    """
    interface, public, preshared, endpoint, ips, handshake, rx, tx, keepalive = fields
    result[interface]["peers"][public] = {
        "available": True,
        "preshared_key": preshared if preshared != "(none)" else "",
        "preshared_key_hash": hash_preshared_key(preshared),
        "endpoint": endpoint if endpoint != "(none)" else "",
        "allowed_ips": ips.split(",") if ips != "(none)" else [],
        "latest_handshake": int(handshake),
        "transfer_rx": int(rx),
        "transfer_tx": int(tx),
        "persistent_keepalive": int(keepalive) if keepalive != "off" else 0,
    }


def derive_public_key(key):
    """Derives the public key from the specified private key.

    Arguments:
        key (str): Private key as a base64 string.

    Returns:
        str: Public key as a base64 string.
    """
    if not key or key == "(none)":
        return ""
    public_key = PrivateKey(key, Base64Encoder).public_key
    return public_key.encode(Base64Encoder).decode("utf-8")


def hash_preshared_key(key):
    """Hashes the specified key.

    Arguments:
        key (str): Base64-encoded key to hash.

    Returns:
        str: Hash of raw key bytes as base64 string.
    """
    if not key or key == "(none)":
        return ""
    return b64encode(sha256(b64decode(key.encode())).digest()).decode()


def separate_dns_and_search(items):
    """Separates the specified list into DNS servers and search domains.

    Arguments:
        items (list): Mixed list of DNS servers and search domains.

    Returns:
        tuple: List of DNS servers, list of search domains.
    """
    dns = []
    search = []
    if items:
        for item in items:
            if re.search("[A-Za-z]", item) and ":" not in item:
                search.append(item)
            else:
                dns.append(item)
    return dns, search


def _get_fwmark_for_default_peer(data):
    """Returns the fwmark for the first interface with peer with a default route.

    Arguments:
        data (dict): Dict parsed from `wg show` command.

    Returns:
        str: Socket mark or blank.
    """
    for name in sorted(data):
        interface = data[name]
        fwmark = interface.get("fwmark")
        if fwmark:
            for peer in interface["peers"].values():
                for ip in peer.get("allowed_ips", []):
                    if ip.endswith("/0"):
                        return fwmark
    return ""



"""cx_Freeze setup script."""

from re import sub

from cx_Freeze import Executable, setup

from procustodibus_agent import DESCRIPTION, DISPLAY_NAME, SERVICE_NAME
from procustodibus_agent import __version__ as version

SERVICE_DEFAULT = "Service"

SERVICE_WIN32_OWN_PROCESS = 0x10
SERVICE_AUTO_START = 0x2
SERVICE_ERROR_NORMAL = 0x1

MSIDB_SERVICE_CONTROL_EVENT_START = 0x1
MSIDB_SERVICE_CONTROL_EVENT_STOP = 0x2
MSIDB_SERVICE_CONTROL_EVENT_UNINSTALL_STOP = 0x20
MSIDB_SERVICE_CONTROL_EVENT_UNINSTALL_DELETE = 0x80

EXECUTABLES = [
    Executable(
        script="procustodibus_agent/windows/service_config.py",
        base="Win32Service",
        target_name="ProCustodibusAgentService",
        icon="installer/favicon.ico",
    ),
    Executable(
        script="procustodibus_agent/cli.py",
        base="console",
        target_name="procustodibus-agent",
        icon="installer/favicon.ico",
    ),
    Executable(
        script="procustodibus_agent/credentials.py",
        base="console",
        target_name="procustodibus-credentials",
        icon="installer/favicon.ico",
    ),
    Executable(
        script="procustodibus_agent/mfa/cli.py",
        base="console",
        target_name="procustodibus-mfa",
        icon="installer/favicon.ico",
    ),
]


def _make_component_id(executables, index):
    executable = executables[index]
    component = f"_cx_executable{index}_{executable}"
    return sub(r"[^\w.]", "_", component)


setup(
    name="procustodibus_agent",
    version=version,
    description=DESCRIPTION,
    options={
        "build_exe": {
            "excludes": [
                "test",
                "tkinter",
                "unittest",
            ],
            "includes": [
                "_cffi_backend",
                "cx_Logging",
            ],
            "include_files": [
                ("LICENSE", "LICENSE.txt"),
                ("installer/help.url", "help.url"),
                ("installer/troubleshooting.url", "troubleshooting.url"),
                ("scripts/windows/", "scripts/"),
                ("installer/cnf.txt", "cnf/README.txt"),
                ("installer/log.txt", "log/README.txt"),
            ],
            "include_msvcr": True,
            "packages": [
                "dns",
                "procustodibus_agent",
            ],
        },
        "bdist_msi": {
            "all_users": True,
            "initial_target_dir": f"[ProgramFiles64Folder]{DISPLAY_NAME}",
            "install_icon": "installer/favicon.ico",
            "upgrade_code": "{EBA087D2-BA25-4E0B-A5FB-3DDC676D0D50}",
            "data": {
                "Directory": [
                    (
                        "ProgramMenuFolder",  # ID
                        "TARGETDIR",  # DirectoryParent
                        ".",  # DefaultDir
                    ),
                    (
                        f"{SERVICE_NAME}Folder",  # ID
                        "ProgramMenuFolder",  # DirectoryParent
                        DISPLAY_NAME,  # DefaultDir
                    ),
                ],
                "Property": [
                    ("cmd", "cmd"),
                    ("explorer", "explorer"),
                ],
                "ServiceInstall": [
                    (
                        f"{SERVICE_NAME}{SERVICE_DEFAULT}Install",  # ID
                        f"{SERVICE_NAME}{SERVICE_DEFAULT}",  # Name
                        DISPLAY_NAME,  # DisplayName
                        SERVICE_WIN32_OWN_PROCESS,  # ServiceType
                        SERVICE_AUTO_START,  # StartType
                        SERVICE_ERROR_NORMAL,  # ErrorControl
                        None,  # LoadOrderGroup
                        None,  # Dependencies
                        None,  # StartName
                        None,  # Password
                        None,  # Arguments
                        _make_component_id(EXECUTABLES, 0),  # Component
                        DESCRIPTION,  # Description
                    ),
                ],
                "ServiceControl": [
                    (
                        f"{SERVICE_NAME}{SERVICE_DEFAULT}Control",  # ID
                        f"{SERVICE_NAME}{SERVICE_DEFAULT}",  # Name
                        (
                            MSIDB_SERVICE_CONTROL_EVENT_START
                            + MSIDB_SERVICE_CONTROL_EVENT_STOP
                            + MSIDB_SERVICE_CONTROL_EVENT_UNINSTALL_STOP
                            + MSIDB_SERVICE_CONTROL_EVENT_UNINSTALL_DELETE
                        ),  # Event
                        None,  # Arguments
                        0,  # Wait
                        _make_component_id(EXECUTABLES, 0),  # Component
                    ),
                ],
                "Shortcut": [
                    (
                        f"{SERVICE_NAME}TestMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        f"{DISPLAY_NAME} Test",  # Name
                        "TARGETDIR",  # Component
                        "[cmd]",  # Target
                        "/k procustodibus-agent.exe --test",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}VersionMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        f"{DISPLAY_NAME} Version",  # Name
                        "TARGETDIR",  # Component
                        "[cmd]",  # Target
                        "/k procustodibus-agent.exe --version",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}MfaMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        f"{DISPLAY_NAME} MFA Status",  # Name
                        "TARGETDIR",  # Component
                        "[cmd]",  # Target
                        "/k procustodibus-mfa.exe",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}ServiceStatusMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        f"{DISPLAY_NAME} Service Status",  # Name
                        "TARGETDIR",  # Component
                        "[cmd]",  # Target
                        f"/k sc query {SERVICE_NAME}{SERVICE_DEFAULT}",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}ServiceStartMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        f"{DISPLAY_NAME} Service Start",  # Name
                        "TARGETDIR",  # Component
                        "[cmd]",  # Target
                        f"/k sc start {SERVICE_NAME}{SERVICE_DEFAULT}",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}ServiceStopMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        f"{DISPLAY_NAME} Service Stop",  # Name
                        "TARGETDIR",  # Component
                        "[cmd]",  # Target
                        f"/k sc stop {SERVICE_NAME}{SERVICE_DEFAULT}",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}HelpMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        "Help",  # Name
                        "TARGETDIR",  # Component
                        "[TARGETDIR]help.url",  # Target
                        None,  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}TroubleshootingMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        "Troubleshooting",  # Name
                        "TARGETDIR",  # Component
                        "[TARGETDIR]troubleshooting.url",  # Target
                        None,  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}ConfigurationMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        "Configuration",  # Name
                        "TARGETDIR",  # Component
                        "[explorer]",  # Target
                        "cnf",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                    (
                        f"{SERVICE_NAME}LogsMenuItem",  # ID
                        f"{SERVICE_NAME}Folder",  # Directory
                        "Logs",  # Name
                        "TARGETDIR",  # Component
                        "[explorer]",  # Target
                        "log",  # Arguments
                        None,  # Description
                        None,  # Hotkey
                        None,  # Icon
                        None,  # IconIndex
                        None,  # ShowCmd
                        "TARGETDIR",  # WkDir
                    ),
                ],
            },
        },
    },
    executables=EXECUTABLES,
)



"""Connectivity test."""

import os
import sys
from re import search
from socket import AF_INET, AF_INET6

from dns.exception import DNSException
from requests import RequestException

from procustodibus_agent import DOCS_URL
from procustodibus_agent.api import (
    get_health_info,
    get_host_info,
    raise_unless_has_cnf,
    setup_api,
)
from procustodibus_agent.resolve_hostname import get_resolver, is_likely_ip
from procustodibus_agent.wg import parse_wg_show, run_wg_show, update_socket_mark
from procustodibus_agent.wg_cnf import load_all_from_wg_cnf


def check_connectivity(cnf, output=None):
    """Runs all connectivity checks and outputs issues.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to (defaults to stdout).

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    if not output:
        output = sys.stdout

    try:
        raise_unless_has_cnf(cnf)
    except ValueError as e:
        print(str(e), file=output)
        return 1

    exit_code = (
        check_wg(cnf, output)
        + check_dns(cnf, output)
        + check_health(cnf, output)
        + check_host(cnf, output)
    )

    if exit_code:
        print(
            f"Issues encountered; see {DOCS_URL}/guide/agents/troubleshoot/ to fix",
            file=output,
        )
    else:
        print("All systems go :)", file=output)

    return exit_code


def check_wg(cnf, output):
    """Checks that wireguard is available and configured with at least one interface.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    if cnf.wiresock:
        try:
            interfaces = load_all_from_wg_cnf(cnf)
        except Exception as e:  # noqa: BLE001
            _bad(f"cannot open interface conf files ({e})", output)
            return 2
    else:
        try:
            interfaces = parse_wg_show(run_wg_show(cnf))
            update_socket_mark(interfaces, cnf)
        except OSError as e:
            _bad(f"no wg executable found ({e})", output)
            return 2

    if interfaces:
        _good(f"{len(interfaces)} wireguard interfaces found", output)
        return 0
    _bad("no wireguard interfaces found", output)
    return 0


def check_dns(cnf, output):
    """Checks that the local DNS resolver can resolve api.procustodib.us.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    hostname = _get_hostname(cnf.api)
    if is_likely_ip(hostname):
        version = 6 if ":" in hostname else 4
        _good(f"{hostname} is pro custodibus api ipv{version} address", output)
        return 0

    resolver = get_resolver(cnf)
    families = [AF_INET6, AF_INET] if resolver.prefer_ipv6 else [AF_INET, AF_INET6]

    result = _check_dns_family(resolver, hostname, families[0], output)
    if (
        result == 0
        or not resolver.has_ipv6
        or cnf.resolve_hostnames in ("ipv4", "ipv6")
    ):
        return result

    result = _check_dns_family(resolver, hostname, families[1], output)
    if result == 0:
        resolver.flip_ipv6_preference()
    return result


def _check_dns_family(resolver, hostname, family, output):
    """Checks that the local DNS resolver can resolve api.procustodib.us.

    Arguments:
        resolver (Resolver): Resolver object.
        hostname (str): Hostname to lookup (eg 'foo.example.com').
        family (int): Preferred family (eg AF_INET6).
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    version = 6 if family == AF_INET6 else 4
    try:
        addresses = resolver.lookup_ips(hostname, family)
        if addresses:
            _good(f"{addresses[0]} is pro custodibus api ipv{version} address", output)
            return 0
    except DNSException as e:
        _bad(f"cannot resolve pro custodibus api ipv{version} address: {e}", output)
        return 4
    else:
        _bad(f"no pro custodibus api ipv{version} address", output)
        return 4


def check_health(cnf, output):
    """Checks connectivity to and the health of the Pro Custodibus API.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    try:
        errors = [x["error"] for x in get_health_info(cnf) if not x["healthy"]]
    except RequestException as e:
        errors = _check_health_of_other_family(cnf, output, e)
    except DNSException as e:
        errors = [_build_server_unvailable_message(e)]

    if errors:
        for error in errors:
            _bad(f"unhealthy pro custodibus api: {error}", output)
        return 8
    _good("healthy pro custodibus api", output)
    return 0


def _check_health_of_other_family(cnf, output, error):
    """Checks connectivity to the API using the other IP version (4 or 6).

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.
        error: Error that triggered this check.

    Returns:
        list: List of errors messages (or empty if no errors).
    """
    resolver = get_resolver(cnf)
    hostname = _get_hostname(cnf.api)
    if (
        not resolver.has_ipv6
        or cnf.resolve_hostnames in ("ipv4", "ipv6")
        or is_likely_ip(hostname)
    ):
        return [_build_server_unvailable_message(error)]

    _bad(_build_server_unvailable_message(error, resolver), output)
    resolver.flip_ipv6_preference()
    family = AF_INET6 if resolver.prefer_ipv6 else AF_INET
    if _check_dns_family(resolver, hostname, family, output):
        return [_build_server_unvailable_message()]

    try:
        return [x["error"] for x in get_health_info(cnf) if not x["healthy"]]
    except (RequestException, DNSException) as e:
        return [_build_server_unvailable_message(e, resolver)]


def _build_server_unvailable_message(error=None, resolver=None):
    """Builds the 'server unavailable' error message.

    Arguments:
        error: Error that triggered this check (or none).
        resolver: DNS resolver (or none), to display IP version (4 or 6).

    Returns:
        str: Error message (eg 'server unavailable').
    """
    if not error:
        return "server unavailable"
    if not resolver:
        return f"server unavailable ({error})"
    version = 6 if resolver.prefer_ipv6 else 4
    return f"server unavailable at ipv{version} address ({error})"


def check_host(cnf, output):
    """Checks that the agent can access the configured host through the API.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    try:
        _setup_if_available(cnf)
    except (DNSException, RequestException, ValueError) as e:
        _bad(f"cannot set up access to api ({e})", output)
        return 16

    try:
        host = get_host_info(cnf)
        name = host["data"][0]["attributes"]["name"]
        _good(f"can access host record on api for {name}", output)
    except (DNSException, RequestException, ValueError) as e:
        _bad(f"cannot access host record on api ({e})", output)
        return 16
    else:
        return 0


def _setup_if_available(cnf):
    """Sets up new agent credentials if setup code is available.

    Arguments:
        cnf (Config): Config object.
    """
    if type(cnf.setup) is dict or os.path.exists(cnf.setup):
        setup_api(cnf)


def _good(message, output):
    """Prints the specified "good" message to the specified output stream.

    Arguments:
        message (str): Message to print.
        output (IOBase): Output stream to write to.
    """
    print(f"... {message} ...", file=output)


def _bad(message, output):
    """Prints the specified "bad" message to the specified output stream.

    Arguments:
        message (str): Message to print.
        output (IOBase): Output stream to write to.
    """
    print(f"!!! {message} !!!", file=output)


def _get_hostname(url):
    """Extracts the hostname from the specified URL.

    Arguments:
        url (str): URL (eg 'http://test.example.com:8080').

    Returns:
        str: Hostname (eg 'test.example.com').
    """
    match = search(r"(?<=://)[^:/]+", url)
    return match.group(0) if match else None



"""cx_Freeze config for the agent as a Win32Service."""

import procustodibus_agent as agent

NAME = f"{agent.SERVICE_NAME}%s"
DISPLAY_NAME = f"{agent.DISPLAY_NAME} %s"
MODULE_NAME = "procustodibus_agent.windows.service"
CLASS_NAME = "Service"
DESCRIPTION = agent.DESCRIPTION
AUTO_START = False
SESSION_CHANGES = False
